import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Frame;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Scanner;
import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import javax.swing.WindowConstants;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;

/**
 * Setup - a class that implements a setup dialog for experimental applications in Java.
 * <p>
 * 
 * A setup dialog is used to load parameters for an application and to present the parameters in a
 * UI that pops up before the application launches. The user is given the opportunity to change
 * setup parameters as appropriate for the current launch of the application. Changes to parameters
 * may be saved and used as default values for the next launch.
 * <p>
 * 
 * Five classes are defined herein: <code>Setup</code>, <code>SetupItem</code>,
 * <code>SetupItemInfo</code>, <code>SetupEvent</code>, <code>Configuration</code>. There is also a
 * listener named <code>SetupItemListener</code> defined in <code>SetupItem</code>.
 * <P>
 * 
 * The motivation for the <code>Setup</code> class is to provide a single, consistent setup mechanism
 * for use across different experimental applications. Examples are seen in the APIs for
 * GraffitiExperiment, H4WriterExperiment, LetterGuessingExperiment, PhoneKeypadExperiment,
 * SAKExperiment, SoftKeyboardExperiment, and TypingTestExperiment. Below is the setup dialog from
 * SoftKeyboardExperiment: (click to enlarge)
 * <p>
 * 
 * <center><a href="Setup-0.jpg"><img src="Setup-0.jpg" width=400></a> </center>
 * <p>
 * 
 * There are four types of setup items that can be combined in designing a setup dialog:
 * <p>
 * 
 * <ul>
 * <li>ComboBox setup item - consists of a <code>JLabel</code> (left) and a <code>JComboBox</code>
 * (right), for example
 * <p>
 * 
 * <center><img src="Setup-1.jpg"> </center>
 * <p>
 * 
 * <li>TextField setup item - consists of a <code>JLabel</code> (left) and a <code>JTextField</code>
 * (right), for example
 * <p>
 * 
 * <center><img src="Setup-2.jpg"> </center>
 * <p>
 * 
 * <li>CheckBox setup item - consists of a <code>JCheckBox</code> (left) and a <code>JLabel</code>
 * (right), for example
 * <p>
 * 
 * <center><img src="Setup-3.jpg"> </center>
 * <p>
 * 
 * <li>RadioButton setup item - consists of a set of <code>JRadioButton</code> objects forming a
 * <code>ButtonGroup</code>, for example
 * <p>
 * 
 * <center> <img src="Setup-4.jpg"> </center>
 * <p>
 * 
 * </ul>
 * 
 * The <code>Setup</code> class extends <code>JDialog</code>:
 * <p>
 * 
 * <pre>
 *      public class Setup extends JDialog implements SetupItem.SetupItemListener, ActionListener
 *      {
 *           ...
 *           Setup(Frame owner, Configuration cArg, String appNameArg, SetupItemInfo[] itemInfoArg)
 *           {
 *                ...
 * </pre>
 * 
 * The constructor requires four arguments: (i) the owner (use <code>null</code>), (ii) an instance
 * of <code>Configuration</code> defining the initial configuration parameters, (iii) the
 * application's name as a string, and (iv) an array of <code>SetupItemInfo</code> objects.
 * <p>
 * 
 * For specific coding details on using <code>Setup</code>, consult the source code for one of the
 * example applications (e.g., <code>SAKExperiment.java</code>).
 * <p>
 * 
 * Note: The example applications do not include a setup parameter for "Block". The block code is
 * generated automatically in the application software. Consult the application source code for
 * details.
 * <p>
 * 
 * @author Scott MacKenzie, 2014
 * 
 */

public class Setup extends JDialog implements SetupItem.SetupItemListener, ActionListener
{
	// the following avoids a "warning" with Java 1.5.0 complier (?)
	static final long serialVersionUID = 42L;

	Configuration c, cSave;
	String appName;
	JLabel banner;

	SetupItem[] items;
	SetupItemInfo[] itemInfo;

	JButton okButton;
	JButton saveButton;
	JButton resetButton;
	JButton exitButton;
	boolean ok;

	final Font BIG = new Font("sansserif", Font.PLAIN, 16);

	/**
	 * Setup - a UI for presenting and changing parameters that set the operation of an application.
	 * 
	 * @param owner
	 *            null
	 * @param cArg
	 *            a <code>Configuration</code> object holding the initial parameter values for the
	 *            setup dialog
	 * @param appNameArg
	 *            a <code>String</code> holding the name of the application
	 * @param itemInfoArg
	 *            an array of <code>SetupItemInfo</code> objects that define the entries in the
	 *            setup dialog
	 */
	public Setup(Frame owner, Configuration cArg, String appNameArg, SetupItemInfo[] itemInfoArg)
	{
		super(owner, "Configure " + appNameArg, true);
		c = cArg;
		appName = appNameArg;
		itemInfo = itemInfoArg;
		cSave = new Configuration(appName, c.toArray());

		setResizable(false);
		setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);

		banner = new JLabel("Setup", SwingConstants.CENTER);
		banner.setFont(new Font("sansserif", Font.PLAIN, 24));
		banner.setBorder(BorderFactory.createEmptyBorder(30, 30, 30, 30));

		/*
		 * Each entry in the setup dialog is defined by a SetupItem. Here, we instantiate a
		 * SetupItem array and populate it with SetupItem instances. Each instance is constructed
		 * using information in the SetupItemInfo array which was passed in to the Setup
		 * constructor.
		 */
		items = new SetupItem[itemInfo.length];
		for (int i = 0; i < itemInfo.length; ++i)
		{
			items[i] = new SetupItem(itemInfo[i].itemType, itemInfo[i].itemLabel, itemInfo[i].itemValues);
			items[i].addSetupItemListener(this);
		}

		okButton = new JButton("OK");
		okButton.setFont(BIG);
		okButton.addActionListener(this);

		saveButton = new JButton("Save");
		saveButton.setFont(BIG);
		saveButton.addActionListener(this);

		resetButton = new JButton("Reset");
		resetButton.setFont(BIG);
		resetButton.addActionListener(this);

		exitButton = new JButton("Exit");
		exitButton.setFont(BIG);
		exitButton.addActionListener(this);

		okButton.setPreferredSize(resetButton.getPreferredSize());
		saveButton.setPreferredSize(resetButton.getPreferredSize());
		exitButton.setPreferredSize(resetButton.getPreferredSize());

		setDefaults();

		JPanel top = new JPanel();
		top.setLayout(new BorderLayout());

		// add items to parameter panel and determine/set the item widths for a consistent look
		int preferredLeftWidth = 0;
		int preferredRightWidth = 0;
		JPanel paramPanel = new JPanel();
		paramPanel.setLayout(new BoxLayout(paramPanel, BoxLayout.Y_AXIS));
		for (int i = 0; i < items.length; ++i)
		{
			paramPanel.add(items[i]);
			if (items[i].getPreferredLeftWidth() > preferredLeftWidth)
				preferredLeftWidth = items[i].getPreferredLeftWidth();

			if (items[i].getPreferredRightWidth() > preferredRightWidth)
				preferredRightWidth = items[i].getPreferredRightWidth();
		}

		for (int i = 0; i < items.length; ++i)
			items[i].setPreferredItemWidth(preferredLeftWidth, preferredRightWidth);

		paramPanel.setBorder(new TitledBorder(new EtchedBorder(), "Parameters"));

		JPanel buttonPanel = new JPanel();
		buttonPanel.add(okButton);
		buttonPanel.add(saveButton);
		buttonPanel.add(resetButton);
		buttonPanel.add(exitButton);
		buttonPanel.setBorder(BorderFactory.createEmptyBorder(30, 30, 30, 30));

		top.add("North", banner);
		top.add("Center", paramPanel);
		top.add("South", buttonPanel);
		top.add("West", new JLabel("            "));
		top.add("East", new JLabel("            "));
		this.setContentPane(top);
		this.pack();
		okButton.requestFocus();
	}

	public void actionPerformed(ActionEvent ae)
	{
		Object source = ae.getSource();

		if (source == resetButton)
			setDefaults();

		else if (source == saveButton)
			saveToFile();

		else if (source == okButton)
		{
			ok = true;
			this.setVisible(false);
		}

		else if (source == exitButton)
			System.exit(0);

		return;
	}

	public void onSetupItemChanged(SetupEvent e)
	{
		int idx;
		String value = null;
		for (idx = 0; idx < items.length; ++idx)
		{
			Object o1 = items[idx].getParameterObject();
			Object o2 = e.getSource();
			if (o1 == o2)
				break;
		}

		if (idx >= items.length)
			return;

		if (items[idx].type == SetupItem.CHECK_BOX)
			value = Boolean.toString(items[idx].setupCheckBox.isSelected());

		else if (items[idx].type == SetupItem.COMBO_BOX)
			value = (String)items[idx].setupComboBox.getSelectedItem();

		else if (items[idx].type == SetupItem.TEXT_FIELD)
			value = items[idx].setupTextField.getText();

		else if (items[idx].type == SetupItem.RADIO_BUTTON)
		{
			for (int i = 0; i < items[idx].setupRadioButtons.length; ++i)
			{
				if (items[idx].setupRadioButtons[i].isSelected())
				{
					value = items[idx].setupRadioButtons[i].getText();
					break;
				}
			}
		}

		c.setConfigurationParameter(idx, value);
	}

	private void setDefaults()
	{
		for (int i = 0; i < items.length; ++i)
			items[i].setSelected(items[i].type, cSave.getConfigurationParameter(i));
	}

	private void saveToFile()
	{
		PrintWriter out;
		try
		{
			out = new PrintWriter(new BufferedWriter(new FileWriter(appName + ".cfg")));
			out.print(c.toString());
			out.flush();
			out.close();
		} catch (IOException ioe)
		{
			JOptionPane.showMessageDialog(null, "Unable to save config to " + appName + ".cfg\n"
					+ "Please ensure the file is writable.", "I/O Error", JOptionPane.ERROR_MESSAGE);
		}
	}

	public boolean showSetup(Frame f)
	{
		ok = false;
		this.setLocationRelativeTo(f);
		this.setVisible(true);
		return ok;
	}
}

// ---------------------
// S E T U P _ E V E N T
// ---------------------

class SetupEvent
{
	Object source;

	SetupEvent(Object sourceArg)
	{
		source = sourceArg;
	}

	Object getSource()
	{
		return source;
	}
}

// -----------------------------
// S E T U P _ I T E M _ I N F O
// -----------------------------

class SetupItemInfo
{
	int itemType;
	String itemLabel;
	String[] itemValues;

	SetupItemInfo(int itemTypeArg, String itemLabelArg, String[] itemValuesArg)
	{
		itemType = itemTypeArg;
		itemLabel = itemLabelArg;
		itemValues = itemValuesArg;
	}
}

// -------------------
// S E T U P _ I T E M
// -------------------

class SetupItem extends JPanel implements ActionListener, ItemListener, FocusListener
{
	// the following avoids a "warning" with Java 1.5.0 complier
	private static final long serialVersionUID = 1L;

	final static int COMBO_BOX = 100;
	final static int TEXT_FIELD = 200;
	final static int CHECK_BOX = 300;
	final static int RADIO_BUTTON = 400;

	final Font BIG = new Font("sansserif", Font.PLAIN, 16);

	JLabel comboBoxLabel;
	JComboBox<String> setupComboBox;

	JLabel textFieldLabel;
	JTextField setupTextField;

	JCheckBox setupCheckBox;
	JLabel checkBoxLabel;

	JRadioButton[] setupRadioButtons;
	JLabel radioButtonsLabel;
	JPanel setupRadioButtonPanel;

	int type;
	String label;
	String[] value;
	int preferredLeftWidth, preferredRightWidth;

	SetupItem(int typeArg, String labelArg, String[] valueArg)
	{
		type = typeArg;
		label = labelArg;
		value = valueArg;

		if (type == COMBO_BOX)
		{
			comboBoxLabel = new JLabel(label);
			comboBoxLabel.setFont(BIG);
			comboBoxLabel.setHorizontalAlignment(SwingConstants.RIGHT);
			preferredLeftWidth = comboBoxLabel.getPreferredSize().width;
			this.add(comboBoxLabel);

			setupComboBox = new JComboBox<String>(value);
			setupComboBox.setFont(BIG);
			setupComboBox.addActionListener(this);
			preferredRightWidth = setupComboBox.getPreferredSize().width;
			this.add(setupComboBox);
		}

		else if (type == TEXT_FIELD)
		{
			textFieldLabel = new JLabel(label);
			textFieldLabel.setFont(BIG);
			textFieldLabel.setHorizontalAlignment(SwingConstants.RIGHT);
			preferredLeftWidth = textFieldLabel.getPreferredSize().width;
			this.add(textFieldLabel);

			setupTextField = new JTextField(value[0]);
			setupTextField.setFont(BIG);
			setupTextField.addActionListener(this);
			setupTextField.addFocusListener(this);
			preferredRightWidth = setupTextField.getPreferredSize().width;
			this.add(setupTextField);
		}

		else if (type == CHECK_BOX)
		{
			setupCheckBox = new JCheckBox();
			setupCheckBox.setFont(BIG);
			setupCheckBox.addItemListener(this);
			setupCheckBox.setHorizontalAlignment(SwingConstants.RIGHT);
			preferredLeftWidth = setupCheckBox.getPreferredSize().width;
			this.add(setupCheckBox);

			checkBoxLabel = new JLabel(label);
			checkBoxLabel.setFont(BIG);
			checkBoxLabel.setHorizontalAlignment(SwingConstants.LEFT);
			preferredRightWidth = checkBoxLabel.getPreferredSize().width;
			this.add(checkBoxLabel);
		}

		else if (type == RADIO_BUTTON)
		{
			setupRadioButtonPanel = new JPanel();
			setupRadioButtonPanel.setBorder(new TitledBorder(new EtchedBorder(), label));

			if (value.length < 4)
				setupRadioButtonPanel.setLayout(new GridLayout(1, 0)); // horizontal
			else
				setupRadioButtonPanel.setLayout(new GridLayout(0, 1)); // vertical

			ButtonGroup bg = new ButtonGroup();

			setupRadioButtons = new JRadioButton[value.length];
			for (int i = 0; i < setupRadioButtons.length; ++i)
			{
				setupRadioButtons[i] = new JRadioButton(value[i]);
				setupRadioButtons[i].setFont(BIG);
				setupRadioButtons[i].addActionListener(this);
				setupRadioButtonPanel.add(setupRadioButtons[i]);
				bg.add(setupRadioButtons[i]);
			}
			setupRadioButtons[0].setSelected(true);
			this.add(setupRadioButtonPanel);
		}
	}

	public int getPreferredLeftWidth()
	{
		return preferredLeftWidth;
	}

	public int getPreferredRightWidth()
	{
		return preferredRightWidth;
	}

	public void setPreferredItemWidth(int left, int right)
	{
		switch (this.type)
		{
			case SetupItem.CHECK_BOX:
				setupCheckBox.setPreferredSize(new Dimension(left, setupCheckBox.getPreferredSize().height));
				checkBoxLabel.setPreferredSize(new Dimension(right, checkBoxLabel.getPreferredSize().height));
				break;
			case SetupItem.COMBO_BOX:

				comboBoxLabel.setPreferredSize(new Dimension(left, comboBoxLabel.getPreferredSize().height));
				setupComboBox.setPreferredSize(new Dimension(right, setupComboBox.getPreferredSize().height));
				break;
			case SetupItem.TEXT_FIELD:

				textFieldLabel.setPreferredSize(new Dimension(left, textFieldLabel.getPreferredSize().height));
				setupTextField.setPreferredSize(new Dimension(right, setupTextField.getPreferredSize().height));
				break;

			case SetupItem.RADIO_BUTTON:
				setupRadioButtonPanel.setPreferredSize(new Dimension(left + right, setupRadioButtonPanel
						.getPreferredSize().height));
		}
	}

	public static interface SetupItemListener
	{
		public abstract void onSetupItemChanged(SetupEvent se);
	}

	SetupItemListener setupItemListener;

	public void addSetupItemListener(SetupItemListener setupItemListenerArg)
	{
		setupItemListener = setupItemListenerArg;
	}

	@Override
	public void focusGained(FocusEvent arg0)
	{
	}

	@Override
	public void focusLost(FocusEvent arg0)
	{
		setupItemListener.onSetupItemChanged(new SetupEvent(this.getParameterObject()));
	}

	@Override
	public void itemStateChanged(ItemEvent arg0)
	{
		setupItemListener.onSetupItemChanged(new SetupEvent(this.getParameterObject()));
	}

	@Override
	public void actionPerformed(ActionEvent ae)
	{
		setupItemListener.onSetupItemChanged(new SetupEvent(this.getParameterObject()));
	}

	public void setSelected(int typeArg, String newValueArg)
	{
		int type = typeArg;
		String newValue = newValueArg;

		if (type == SetupItem.CHECK_BOX)
		{
			setupCheckBox.setSelected(Boolean.valueOf(newValue));
		}

		else if (type == SetupItem.COMBO_BOX)
		{
			for (int i = 0; i < value.length; ++i)
			{
				if (value[i].equals(newValue))
				{
					setupComboBox.setSelectedIndex(i);
					break;
				}
			}
		}

		else if (type == SetupItem.TEXT_FIELD)
		{
			setupTextField.setText(newValue);
		}

		else if (type == SetupItem.RADIO_BUTTON)
		{
			for (int i = 0; i < setupRadioButtons.length; ++i)
			{
				if (setupRadioButtons[i].getText().equals(newValue))
				{
					setupRadioButtons[i].setSelected(true);
					break;
				}
			}
		}
	}

	public Object getParameterObject()
	{
		Object o = null;
		if (type == SetupItem.CHECK_BOX)
			o = setupCheckBox;
		else if (type == SetupItem.COMBO_BOX)
			o = setupComboBox;
		else if (type == SetupItem.TEXT_FIELD)
			o = setupTextField;
		else if (type == SetupItem.RADIO_BUTTON)
		{
			for (int i = 0; i < setupRadioButtons.length; ++i)
				if (setupRadioButtons[i].isSelected())
				{
					o = setupRadioButtons[i];
					break;
				}
		}
		return o;
	}
}

// -------------------------
// C O N F I G U R A T I O N
// -------------------------

class Configuration
{
	String appName;
	String[] parameters;

	Configuration(String appNameArg, String[] parametersArg)
	{
		appName = appNameArg;
		parameters = new String[parametersArg.length];
		System.arraycopy(parametersArg, 0, parameters, 0, parametersArg.length);
	}

	// ----------
	// set methods
	// -----------

	void setConfigurationParameter(int idx, String par)
	{
		parameters[idx] = par;
	}

	// -----------
	// get methods
	// -----------

	String getConfigurationParameter(int idx)
	{
		return parameters[idx];
	}

	String[] toArray()
	{
		return parameters;
	}

	public static Configuration readConfigurationData(String appNameArg, String[] defaultParametersArg,
			String[] otherFilesArg) throws IOException
	{
		String appName = appNameArg;
		String[] defaultParameters = defaultParametersArg;
		String[] otherFiles = otherFilesArg; // other files to copy (may be null)

		BufferedReader br = null;
		String fileName = appName + ".cfg"; // name of the configuration file
		String line = "";

		// Added for executable Jar:
		File cfgFile = new File(fileName);

		if (!cfgFile.exists())
		{
			String msg = "The " + fileName + " file was not found. Would you\n"
					+ "like a default configuration file to be created for you?";
			int resp = JOptionPane
					.showConfirmDialog(null, msg, "Missing Configuration File", JOptionPane.YES_NO_OPTION);
			if (resp == JOptionPane.YES_OPTION)
			{
				PrintWriter out;
				try
				{
					out = new PrintWriter(new BufferedWriter(new FileWriter(appName + ".cfg")));
					out.print((new Configuration(appName, defaultParameters)).toString());
					out.flush();
					out.close();
				} catch (IOException ioe)
				{
					JOptionPane.showMessageDialog(null, "Unable to save config to " + appName + ".cfg\n"
							+ "Please ensure the file is writable.", "I/O Error", JOptionPane.ERROR_MESSAGE);
				}
			} else
				System.exit(0);
		}

		// needed when executing from .jar file for the first time
		for (int i = 0; otherFiles != null && i < otherFiles.length; ++i)
		{
			if (!(new File(otherFiles[i]).exists()))
				copyFile(appName, otherFiles[i]);
		}

		// with the above preamble (maybe), we can now read the .cfg file
		try
		{
			br = new BufferedReader(new FileReader(cfgFile));
		} catch (FileNotFoundException e)
		{
			showError("Configuration file not found: " + fileName);
			System.exit(1);
		}

		// ----------------------------------
		// read lines from configuration file
		// ----------------------------------

		ArrayList<String> tmp = new ArrayList<String>();

		// skip over blank lines and lines beginning with "#", then read the next parameter
		while (true)
		{
			while ((line = br.readLine()) != null && (line.trim().length() == 0 || line.charAt(0) == '#'))
				continue;
			if (line == null)
				break;
			else
				tmp.add(line);
		} // finished reading configuration file

		String[] parameters = new String[tmp.size()];
		parameters = tmp.toArray(parameters);
		Configuration c = new Configuration(appName, parameters);
		return c;
	}

	private static void showError(String msg)
	{
		JOptionPane.showMessageDialog(null, msg, "I/O Error", JOptionPane.ERROR_MESSAGE);
	}

	// copy a resource file (in the Jar) to the working directory
	private static void copyFile(String appNameArg, String fileNameArg) throws IOException
	{
		String appName = appNameArg;
		String fileName = fileNameArg;

		// use a 'generic' approach to copy a resource file
		Class<?> appClass = null;
		try
		{
			appClass = Class.forName(appName);
		} catch (ClassNotFoundException e)
		{
			e.printStackTrace();
		}
		Scanner in = new Scanner(appClass.getResourceAsStream(fileName));

		PrintStream out = new PrintStream(fileName);
		while (in.hasNextLine())
		{
			out.println(in.nextLine());
		}
		out.flush();
		out.close();
		in.close();
	}

	/*
	 * Return the configuration parameters as a string. The string is written to the .cfg file. The
	 * parameters passed in to this method are only used the first time the method is called (i.e.,
	 * when launching the application for the first time). Normally, this method is called when the
	 * user clicks the Save button in the setup dialog's UI. In this case, the current values of the
	 * setup parameters are returned in the configuration string.
	 */
	public String toString()
	{
		String title = "Configuration parameters for " + appName + "\n";

		StringBuilder sb = new StringBuilder(700);
		sb.append("# ");
		for (int i = 0; i < title.length() - 1; ++i)
			sb.append("="); // ================ (as needed)
		sb.append("\n");
		sb.append("# " + title);
		sb.append("# ");
		for (int i = 0; i < title.length() - 1; ++i)
			sb.append("="); // ================ (as needed)
		sb.append("\n");
		sb.append("\n");

		for (int i = 0; i < parameters.length; ++i)
		{
			sb.append("# -----\n");
			sb.append("# Parameter #" + (i + 1) + " (as per setup dialog)\n");
			sb.append(parameters[i] + "\n");
			sb.append("\n");
		}
		sb.append("# --- end ---\n");
		return sb.toString();
	}
}
